@@ -1,26 +1,6 @@
-/* PipeWire
- *
- * Copyright © 2018 Wim Taymans
- *
- * Permission is hereby granted, free of charge, to any person obtaining a
- * copy of this software and associated documentation files (the "Software"),
- * to deal in the Software without restriction, including without limitation
- * the rights to use, copy, modify, merge, publish, distribute, sublicense,
- * and/or sell copies of the Software, and to permit persons to whom the
- * Software is furnished to do so, subject to the following conditions:
- *
- * The above copyright notice and this permission notice (including the next
- * paragraph) shall be included in all copies or substantial portions of the
- * Software.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
- * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
- * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
- * DEALINGS IN THE SOFTWARE.
- */
+/* PipeWire */
+/* SPDX-FileCopyrightText: Copyright © 2018 Wim Taymans */
+/* SPDX-License-Identifier: MIT */
 
 #include <errno.h>
 #include <string.h>
@@ -31,14 +11,12 @@
 #include <spa/pod/parser.h>
 #include <spa/pod/compare.h>
 #include <spa/param/param.h>
+#include <spa/debug/types.h>
 
+#define PW_API_LINK_IMPL	SPA_EXPORT
 #include "pipewire/impl-link.h"
 #include "pipewire/private.h"
 
-#include <spa/debug/node.h>
-#include <spa/debug/pod.h>
-#include <spa/debug/format.h>
-
 PW_LOG_TOPIC_EXTERN(log_link);
 #define PW_LOG_TOPIC_DEFAULT log_link
 
@@ -50,13 +28,16 @@ PW_LOG_TOPIC_EXTERN(log_link);
 struct impl {
 	struct pw_impl_link this;
 
-	unsigned int io_set:1;
 	unsigned int activated:1;
 
 	struct pw_work_queue *work;
 
 	uint32_t output_busy_id;
 	uint32_t input_busy_id;
+	int output_pending_seq;
+	int input_pending_seq;
+	int output_result;
+	int input_result;
 
 	struct spa_pod *format_filter;
 	struct pw_properties *properties;
@@ -68,9 +49,10 @@ struct impl {
 	struct spa_hook output_node_listener;
 	struct spa_hook output_global_listener;
 
-	struct spa_io_buffers io;
+	struct spa_io_buffers io[2];
 
 	struct pw_impl_node *inode, *onode;
+	bool async;
 };
 
 /** \endcond */
@@ -91,6 +73,38 @@ static void info_changed(struct pw_impl_link *link)
 	link->info.change_mask = 0;
 }
 
+static inline int input_set_busy_id(struct pw_impl_link *link, uint32_t id, int pending_seq)
+{
+	struct impl *impl = SPA_CONTAINER_OF(link, struct impl, this);
+	int res = impl->input_result;
+	if (impl->input_busy_id != SPA_ID_INVALID)
+		link->input->busy_count--;
+	if (id != SPA_ID_INVALID)
+		link->input->busy_count++;
+	impl->input_busy_id = id;
+	impl->input_pending_seq = SPA_RESULT_ASYNC_SEQ(pending_seq);
+	impl->input_result = 0;
+	if (link->input->busy_count < 0)
+		pw_log_error("%s: invalid busy count:%d", link->name, link->input->busy_count);
+	return res;
+}
+
+static inline int output_set_busy_id(struct pw_impl_link *link, uint32_t id, int pending_seq)
+{
+	struct impl *impl = SPA_CONTAINER_OF(link, struct impl, this);
+	int res = impl->output_result;
+	if (impl->output_busy_id != SPA_ID_INVALID)
+		link->output->busy_count--;
+	if (id != SPA_ID_INVALID)
+		link->output->busy_count++;
+	impl->output_busy_id = id;
+	impl->output_pending_seq = SPA_RESULT_ASYNC_SEQ(pending_seq);
+	impl->output_result = 0;
+	if (link->output->busy_count < 0)
+		pw_log_error("%s: invalid busy count:%d", link->name, link->output->busy_count);
+	return res;
+}
+
 static void link_update_state(struct pw_impl_link *link, enum pw_link_state state, int res, char *error)
 {
 	struct impl *impl = SPA_CONTAINER_OF(link, struct impl, this);
@@ -108,12 +122,16 @@ static void link_update_state(struct pw_impl_link *link, enum pw_link_state stat
 		     pw_link_state_as_string(state), error);
 
 	if (state == PW_LINK_STATE_ERROR) {
-		pw_log_error("(%s) %s -> error (%s)", link->name,
-				pw_link_state_as_string(old), error);
+		pw_log_error("(%s) %s -> error (%s) (%s-%s)", link->name,
+				pw_link_state_as_string(old), error,
+				pw_impl_port_state_as_string(link->output->state),
+				pw_impl_port_state_as_string(link->input->state));
 	} else {
-		pw_log_info("(%s) %s -> %s", link->name,
+		pw_log_info("(%s) %s -> %s (%s-%s)", link->name,
 				pw_link_state_as_string(old),
-				pw_link_state_as_string(state));
+				pw_link_state_as_string(state),
+				pw_impl_port_state_as_string(link->output->state),
+				pw_impl_port_state_as_string(link->input->state));
 	}
 
 	pw_impl_link_emit_state_changed(link, old, state, error);
@@ -134,22 +152,18 @@ static void link_update_state(struct pw_impl_link *link, enum pw_link_state stat
 		link->prepared = true;
 		link->preparing = false;
 		pw_context_recalc_graph(link->context, "link prepared");
-	} else if (old == PW_LINK_STATE_PAUSED && state < PW_LINK_STATE_PAUSED) {
+	} else if (old >= PW_LINK_STATE_PAUSED && state < PW_LINK_STATE_PAUSED) {
 		link->prepared = false;
 		link->preparing = false;
 		pw_context_recalc_graph(link->context, "link unprepared");
 	} else if (state == PW_LINK_STATE_INIT) {
 		link->prepared = false;
 		link->preparing = false;
-		if (impl->output_busy_id != SPA_ID_INVALID) {
-			impl->output_busy_id = SPA_ID_INVALID;
-			link->output->busy_count--;
-		}
+
+		output_set_busy_id(link, SPA_ID_INVALID, SPA_ID_INVALID);
 		pw_work_queue_cancel(impl->work, &link->output_link, SPA_ID_INVALID);
-		if (impl->input_busy_id != SPA_ID_INVALID) {
-			impl->input_busy_id = SPA_ID_INVALID;
-			link->input->busy_count--;
-		}
+
+		input_set_busy_id(link, SPA_ID_INVALID, SPA_ID_INVALID);
 		pw_work_queue_cancel(impl->work, &link->input_link, SPA_ID_INVALID);
 	}
 }
@@ -165,14 +179,14 @@ static void complete_ready(void *obj, void *data, int res, uint32_t id)
 	else
 		port = this->output;
 
-	if (id == impl->input_busy_id) {
-		impl->input_busy_id = SPA_ID_INVALID;
-		port->busy_count--;
-	} else if (id == impl->output_busy_id) {
-		impl->output_busy_id = SPA_ID_INVALID;
-		port->busy_count--;
-	} else if (id != SPA_ID_INVALID)
-		return;
+	if (id != SPA_ID_INVALID) {
+		if (id == impl->input_busy_id)
+			res = input_set_busy_id(this, SPA_ID_INVALID, SPA_ID_INVALID);
+		else if (id == impl->output_busy_id)
+			res = output_set_busy_id(this, SPA_ID_INVALID, SPA_ID_INVALID);
+		else
+			return;
+	}
 
 	pw_log_debug("%p: obj:%p port %p complete state:%d: %s", this, obj, port,
 			port->state, spa_strerror(res));
@@ -181,13 +195,12 @@ static void complete_ready(void *obj, void *data, int res, uint32_t id)
 		if (port->state < PW_IMPL_PORT_STATE_READY)
 			pw_impl_port_update_state(port, PW_IMPL_PORT_STATE_READY,
 					0, NULL);
+		if (this->input->state >= PW_IMPL_PORT_STATE_READY &&
+		    this->output->state >= PW_IMPL_PORT_STATE_READY)
+			link_update_state(this, PW_LINK_STATE_ALLOCATING, 0, NULL);
 	} else {
-		pw_impl_port_update_state(port, PW_IMPL_PORT_STATE_ERROR,
-				res, spa_aprintf("port error going to READY: %s", spa_strerror(res)));
+		link_update_state(this, PW_LINK_STATE_ERROR, -EIO, strdup("Format negotiation failed"));
 	}
-	if (this->input->state >= PW_IMPL_PORT_STATE_READY &&
-	    this->output->state >= PW_IMPL_PORT_STATE_READY)
-		link_update_state(this, PW_LINK_STATE_ALLOCATING, 0, NULL);
 }
 
 static void complete_paused(void *obj, void *data, int res, uint32_t id)
@@ -205,14 +218,14 @@ static void complete_paused(void *obj, void *data, int res, uint32_t id)
 		mix = &this->rt.out_mix;
 	}
 
-	if (id == impl->input_busy_id) {
-		impl->input_busy_id = SPA_ID_INVALID;
-		port->busy_count--;
-	} else if (id == impl->output_busy_id) {
-		impl->output_busy_id = SPA_ID_INVALID;
-		port->busy_count--;
-	} else if (id != SPA_ID_INVALID)
-		return;
+	if (id != SPA_ID_INVALID) {
+		if (id == impl->input_busy_id)
+			res = input_set_busy_id(this, SPA_ID_INVALID, SPA_ID_INVALID);
+		else if (id == impl->output_busy_id)
+			res = output_set_busy_id(this, SPA_ID_INVALID, SPA_ID_INVALID);
+		else
+			return;
+	}
 
 	pw_log_debug("%p: obj:%p port %p complete state:%d: %s", this, obj, port,
 			port->state, spa_strerror(res));
@@ -222,13 +235,13 @@ static void complete_paused(void *obj, void *data, int res, uint32_t id)
 			pw_impl_port_update_state(port, PW_IMPL_PORT_STATE_PAUSED,
 					0, NULL);
 		mix->have_buffers = true;
+
+		if (this->rt.in_mix.have_buffers && this->rt.out_mix.have_buffers)
+			link_update_state(this, PW_LINK_STATE_PAUSED, 0, NULL);
 	} else {
-		pw_impl_port_update_state(port, PW_IMPL_PORT_STATE_ERROR,
-				res, spa_aprintf("port error going to PAUSED: %s", spa_strerror(res)));
 		mix->have_buffers = false;
+		link_update_state(this, PW_LINK_STATE_ERROR, -EIO, strdup("Buffer allocation failed"));
 	}
-	if (this->rt.in_mix.have_buffers && this->rt.out_mix.have_buffers)
-		link_update_state(this, PW_LINK_STATE_PAUSED, 0, NULL);
 }
 
 static void complete_sync(void *obj, void *data, int res, uint32_t id)
@@ -256,8 +269,10 @@ static int do_negotiate(struct pw_impl_link *this)
 	struct pw_impl_port *input, *output;
 	uint8_t buffer[4096];
 	struct spa_pod_builder b = SPA_POD_BUILDER_INIT(buffer, sizeof(buffer));
-	uint32_t index;
+	uint32_t index, busy_id;
 	uint32_t in_state, out_state;
+	struct spa_node *in_node, *out_node;
+	uint32_t in_port, out_port;
 
 	if (this->info.state >= PW_LINK_STATE_NEGOTIATING)
 		return 0;
@@ -277,12 +292,27 @@ static int do_negotiate(struct pw_impl_link *this)
 
 	input = this->input;
 	output = this->output;
+#if 0
+	in_node = input->mix;
+	in_port = this->rt.in_mix.port.port_id;
+	out_node = output->mix;
+	out_port = this->rt.out_mix.port.port_id;
+#else
+	in_node = input->node->node;
+	in_port = input->port_id;
+	out_node = output->node->node;
+	out_port = output->port_id;
+#endif
 
 	/* find a common format for the ports */
 	if ((res = pw_context_find_format(context,
-					output, input, NULL, 0, NULL,
-					&format, &b, &error)) < 0)
+					output, SPA_ID_INVALID,
+					input, SPA_ID_INVALID,
+					NULL, 0, NULL,
+					&format, &b, &error)) < 0) {
+		format = NULL;
 		goto error;
+	}
 
 	format = spa_pod_copy(format);
 	spa_pod_fixate(format);
@@ -292,8 +322,8 @@ static int do_negotiate(struct pw_impl_link *this)
 	/* if output port had format and is idle, check if it changed. If so, renegotiate */
 	if (out_state > PW_IMPL_PORT_STATE_CONFIGURE && output->node->info.state == PW_NODE_STATE_IDLE) {
 		index = 0;
-		res = spa_node_port_enum_params_sync(output->node->node,
-				output->direction, output->port_id,
+		res = spa_node_port_enum_params_sync(out_node,
+				output->direction, out_port,
 				SPA_PARAM_Format, &index,
 				NULL, &current, &b);
 		switch (res) {
@@ -326,8 +356,8 @@ static int do_negotiate(struct pw_impl_link *this)
 	/* if input port had format and is idle, check if it changed. If so, renegotiate */
 	if (in_state > PW_IMPL_PORT_STATE_CONFIGURE && input->node->info.state == PW_NODE_STATE_IDLE) {
 		index = 0;
-		res = spa_node_port_enum_params_sync(input->node->node,
-				input->direction, input->port_id,
+		res = spa_node_port_enum_params_sync(in_node,
+				input->direction, in_port,
 				SPA_PARAM_Format, &index,
 				NULL, &current, &b);
 		switch (res) {
@@ -375,10 +405,11 @@ static int do_negotiate(struct pw_impl_link *this)
 			goto error;
 		}
 		if (SPA_RESULT_IS_ASYNC(res)) {
-			output->busy_count++;
-			res = spa_node_sync(output->node->node, res);
-			impl->output_busy_id = pw_work_queue_add(impl->work, &this->output_link, res,
+			pw_log_info("output set format %d", res);
+			busy_id = pw_work_queue_add(impl->work, &this->output_link,
+					spa_node_sync(output->node->node, res),
 					complete_ready, this);
+			output_set_busy_id(this, busy_id, res);
 		} else {
 			complete_ready(&this->output_link, this, res, SPA_ID_INVALID);
 		}
@@ -394,10 +425,11 @@ static int do_negotiate(struct pw_impl_link *this)
 			goto error;
 		}
 		if (SPA_RESULT_IS_ASYNC(res2)) {
-			input->busy_count++;
-			res2 = spa_node_sync(input->node->node, res2);
-			impl->input_busy_id = pw_work_queue_add(impl->work, &this->input_link, res2,
+			pw_log_info("input set format %d", res2);
+			busy_id = pw_work_queue_add(impl->work, &this->input_link,
+					spa_node_sync(input->node->node, res2),
 					complete_ready, this);
+			input_set_busy_id(this, busy_id, res2);
 			if (res == 0)
 				res = res2;
 		} else {
@@ -415,12 +447,12 @@ static int do_negotiate(struct pw_impl_link *this)
 	return res;
 
 error:
-	pw_context_debug_port_params(context, input->node->node, input->direction,
-			input->port_id, SPA_PARAM_EnumFormat, res,
-			"input format (%s)", error);
-	pw_context_debug_port_params(context, output->node->node, output->direction,
-			output->port_id, SPA_PARAM_EnumFormat, res,
-			"output format (%s)", error);
+	pw_context_debug_port_params(context, in_node,
+			input->direction, in_port,
+			SPA_PARAM_EnumFormat, res, "input format (%s)", error);
+	pw_context_debug_port_params(context, out_node,
+			output->direction, out_port,
+			SPA_PARAM_EnumFormat, res, "output format (%s)", error);
 	link_update_state(this, PW_LINK_STATE_ERROR, res, error);
 	free(format);
 	return res;
@@ -431,7 +463,6 @@ static int port_set_io(struct pw_impl_link *this, struct pw_impl_port *port, uin
 {
 	int res = 0;
 
-	mix->io = data;
 	pw_log_debug("%p: %s port %p %d.%d set io: %d %p %zd", this,
 			pw_direction_as_string(port->direction),
 			port, port->port_id, mix->port.port_id, id, data, size);
@@ -456,21 +487,22 @@ static void select_io(struct pw_impl_link *this)
 	struct impl *impl = SPA_CONTAINER_OF(this, struct impl, this);
 	struct spa_io_buffers *io;
 
-	io = this->rt.in_mix.io;
+	io = this->rt.in_mix.io_data;
 	if (io == NULL)
-		io = this->rt.out_mix.io;
+		io = this->rt.out_mix.io_data;
 	if (io == NULL)
-		io = &impl->io;
+		io = impl->io;
 
 	this->io = io;
-	*this->io = SPA_IO_BUFFERS_INIT;
+	this->io[0] = SPA_IO_BUFFERS_INIT;
+	this->io[1] = SPA_IO_BUFFERS_INIT;
 }
 
 static int do_allocation(struct pw_impl_link *this)
 {
 	struct impl *impl = SPA_CONTAINER_OF(this, struct impl, this);
 	int res;
-	uint32_t in_flags, out_flags;
+	uint32_t in_flags, out_flags, busy_id;
 	char *error = NULL;
 	struct pw_impl_port *input, *output;
 
@@ -508,22 +540,41 @@ static int do_allocation(struct pw_impl_link *this)
 		this->rt.out_mix.have_buffers = true;
 	} else {
 		uint32_t flags, alloc_flags;
+		struct spa_node *in_node, *out_node;
+		uint32_t in_port, out_port;
 
 		flags = 0;
 		/* always shared buffers for the link */
 		alloc_flags = PW_BUFFERS_FLAG_SHARED;
+		/* always enable async mode */
+		alloc_flags |= PW_BUFFERS_FLAG_ASYNC;
+
 		if (output->node->remote || input->node->remote)
 			alloc_flags |= PW_BUFFERS_FLAG_SHARED_MEM;
 
+		if (output->node->driver)
+			alloc_flags |= PW_BUFFERS_FLAG_IN_PRIORITY;
+
 		/* if output port can alloc buffers, alloc skeleton buffers */
 		if (SPA_FLAG_IS_SET(out_flags, SPA_PORT_FLAG_CAN_ALLOC_BUFFERS)) {
 			SPA_FLAG_SET(alloc_flags, PW_BUFFERS_FLAG_NO_MEM);
 			flags |= SPA_NODE_BUFFERS_FLAG_ALLOC;
 		}
+#if 0
+		in_node = input->mix;
+		in_port = this->rt.in_mix.port.port_id;
+		out_node = output->mix;
+		out_port = this->rt.out_mix.port.port_id;
+#else
+		in_node = input->node->node;
+		in_port = input->port_id;
+		out_node = output->node->node;
+		out_port = output->port_id;
+#endif
 
 		if ((res = pw_buffers_negotiate(this->context, alloc_flags,
-						output->node->node, output->port_id,
-						input->node->node, input->port_id,
+						out_node, out_port,
+						in_node, in_port,
 						&output->buffers)) < 0) {
 			error = spa_aprintf("error alloc buffers: %s", spa_strerror(res));
 			goto error;
@@ -540,10 +591,10 @@ static int do_allocation(struct pw_impl_link *this)
 			goto error_clear;
 		}
 		if (SPA_RESULT_IS_ASYNC(res)) {
-			output->busy_count++;
-			res = spa_node_sync(output->node->node, res);
-			impl->output_busy_id = pw_work_queue_add(impl->work, &this->output_link, res,
+			busy_id = pw_work_queue_add(impl->work, &this->output_link,
+					spa_node_sync(output->node->node, res),
 					complete_paused, this);
+			output_set_busy_id(this, busy_id, res);
 			if (flags & SPA_NODE_BUFFERS_FLAG_ALLOC)
 				return 0;
 		} else {
@@ -563,10 +614,10 @@ static int do_allocation(struct pw_impl_link *this)
 	}
 
 	if (SPA_RESULT_IS_ASYNC(res)) {
-		input->busy_count++;
-		res = spa_node_sync(input->node->node, res);
-		impl->input_busy_id = pw_work_queue_add(impl->work, &this->input_link, res,
+		busy_id = pw_work_queue_add(impl->work, &this->input_link,
+				spa_node_sync(input->node->node, res),
 				complete_paused, this);
+		input_set_busy_id(this, busy_id, res);
 	} else {
 		complete_paused(&this->input_link, this, res, SPA_ID_INVALID);
 	}
@@ -579,66 +630,47 @@ error:
 	return res;
 }
 
-static int
-do_activate_link(struct spa_loop *loop,
-		 bool async, uint32_t seq, const void *data, size_t size, void *user_data)
-{
-	struct pw_impl_link *this = user_data;
-	struct impl *impl = SPA_CONTAINER_OF(this, struct impl, this);
-
-	pw_log_trace("%p: activate", this);
-
-	spa_list_append(&this->output->rt.mix_list, &this->rt.out_mix.rt_link);
-	spa_list_append(&this->input->rt.mix_list, &this->rt.in_mix.rt_link);
-
-	if (impl->inode != impl->onode) {
-		struct pw_node_activation_state *state;
-
-		this->rt.target.activation = impl->inode->rt.activation;
-		spa_list_append(&impl->onode->rt.target_list, &this->rt.target.link);
-
-		state = &this->rt.target.activation->state[0];
-		if (!this->rt.target.active && impl->onode->rt.driver_target.node != NULL) {
-			state->required++;
-			this->rt.target.active = true;
-		}
-
-		pw_log_trace("%p: node:%p state:%p pending:%d/%d", this, impl->inode,
-				state, state->pending, state->required);
-	}
-	return 0;
-}
-
 int pw_impl_link_activate(struct pw_impl_link *this)
 {
 	struct impl *impl = SPA_CONTAINER_OF(this, struct impl, this);
 	int res;
+	uint32_t io_type, io_size;
 
 	pw_log_debug("%p: activate activated:%d state:%s", this, impl->activated,
 			pw_link_state_as_string(this->info.state));
 
-	if (impl->activated || !this->prepared || !impl->inode->active || !impl->onode->active)
+	if (this->destroyed || impl->activated || !this->prepared ||
+		!impl->inode->runnable || !impl->onode->runnable)
 		return 0;
 
-	if (!impl->io_set) {
-		if ((res = port_set_io(this, this->output, SPA_IO_Buffers, this->io,
-				sizeof(struct spa_io_buffers), &this->rt.out_mix)) < 0)
-			return res;
-
-		if ((res = port_set_io(this, this->input, SPA_IO_Buffers, this->io,
-				sizeof(struct spa_io_buffers), &this->rt.in_mix)) < 0)
-			return res;
-		impl->io_set = true;
+	if (impl->async) {
+		io_type = SPA_IO_AsyncBuffers;
+		io_size = sizeof(struct spa_io_async_buffers);
+	} else {
+		io_type = SPA_IO_Buffers;
+		io_size = sizeof(struct spa_io_buffers);
 	}
-	pw_loop_invoke(this->output->node->data_loop,
-	       do_activate_link, SPA_ID_INVALID, NULL, 0, false, this);
+
+	if ((res = port_set_io(this, this->input, io_type, this->io,
+					io_size, &this->rt.in_mix)) < 0)
+		goto error;
+	if ((res = port_set_io(this, this->output, io_type, this->io,
+					io_size, &this->rt.out_mix)) < 0)
+		goto error_clean;
 
 	impl->activated = true;
 	pw_log_info("(%s) activated", this->name);
 	link_update_state(this, PW_LINK_STATE_ACTIVE, 0, NULL);
 
 	return 0;
+
+error_clean:
+	port_set_io(this, this->input, io_type, NULL, 0, &this->rt.in_mix);
+error:
+	pw_log_error("%p: can't activate link: %s", this, spa_strerror(res));
+	return res;
 }
+
 static void check_states(void *obj, void *user_data, int res, uint32_t id)
 {
 	struct pw_impl_link *this = obj;
@@ -686,13 +718,13 @@ static void check_states(void *obj, void *user_data, int res, uint32_t id)
 	}
 
 	if (output->busy_count > 0) {
-		pw_log_debug("%p: output port %p was busy", this, output);
+		pw_log_debug("%p: output port %p was busy %d", this, output, output->busy_count);
 		res = spa_node_sync(output->node->node, 0);
 		pw_work_queue_add(impl->work, &this->output_link, res, complete_sync, this);
 		goto exit;
 	}
 	else if (input->busy_count > 0) {
-		pw_log_debug("%p: input port %p was busy", this, input);
+		pw_log_debug("%p: input port %p was busy %d", this, input, input->busy_count);
 		res = spa_node_sync(input->node->node, 0);
 		pw_work_queue_add(impl->work, &this->input_link, res, complete_sync, this);
 		goto exit;
@@ -722,10 +754,8 @@ static void input_remove(struct pw_impl_link *this, struct pw_impl_port *port)
 
 	pw_log_debug("%p: remove input port %p", this, port);
 
-	if (impl->input_busy_id != SPA_ID_INVALID) {
-		impl->input_busy_id = SPA_ID_INVALID;
-		port->busy_count--;
-	}
+	input_set_busy_id(this, SPA_ID_INVALID, SPA_ID_INVALID);
+
 	spa_hook_remove(&impl->input_port_listener);
 	spa_hook_remove(&impl->input_node_listener);
 	spa_hook_remove(&impl->input_global_listener);
@@ -734,10 +764,13 @@ static void input_remove(struct pw_impl_link *this, struct pw_impl_port *port)
 	pw_impl_port_emit_link_removed(this->input, this);
 
 	pw_impl_port_recalc_latency(this->input);
+	pw_impl_port_recalc_tag(this->input);
 
-	if ((res = pw_impl_port_use_buffers(port, mix, 0, NULL, 0)) < 0) {
+	if ((res = port_set_io(this, this->input, SPA_IO_Buffers, NULL, 0, mix)) < 0)
+		pw_log_warn("%p: port %p set_io error %s", this, port, spa_strerror(res));
+	if ((res = pw_impl_port_use_buffers(port, mix, 0, NULL, 0)) < 0)
 		pw_log_warn("%p: port %p clear error %s", this, port, spa_strerror(res));
-	}
+
 	pw_impl_port_release_mix(port, mix);
 
 	pw_work_queue_cancel(impl->work, &this->input_link, SPA_ID_INVALID);
@@ -751,10 +784,8 @@ static void output_remove(struct pw_impl_link *this, struct pw_impl_port *port)
 
 	pw_log_debug("%p: remove output port %p", this, port);
 
-	if (impl->output_busy_id != SPA_ID_INVALID) {
-		impl->output_busy_id = SPA_ID_INVALID;
-		port->busy_count--;
-	}
+	output_set_busy_id(this, SPA_ID_INVALID, SPA_ID_INVALID);
+
 	spa_hook_remove(&impl->output_port_listener);
 	spa_hook_remove(&impl->output_node_listener);
 	spa_hook_remove(&impl->output_global_listener);
@@ -763,6 +794,7 @@ static void output_remove(struct pw_impl_link *this, struct pw_impl_port *port)
 	pw_impl_port_emit_link_removed(this->output, this);
 
 	pw_impl_port_recalc_latency(this->output);
+	pw_impl_port_recalc_tag(this->output);
 
 	/* we don't clear output buffers when the link goes away. They will get
 	 * cleared when the node goes to suspend */
@@ -776,14 +808,14 @@ int pw_impl_link_prepare(struct pw_impl_link *this)
 {
 	struct impl *impl = SPA_CONTAINER_OF(this, struct impl, this);
 
-	pw_log_debug("%p: prepared:%d preparing:%d in_active:%d out_active:%d",
+	pw_log_debug("%p: prepared:%d preparing:%d in_active:%d out_active:%d passive:%u",
 			this, this->prepared, this->preparing,
-			impl->inode->active, impl->onode->active);
+			impl->inode->active, impl->onode->active, this->passive);
 
 	if (!impl->inode->active || !impl->onode->active)
 		return 0;
 
-	if (this->preparing || this->prepared)
+	if (this->destroyed || this->preparing || this->prepared)
 		return 0;
 
 	this->preparing = true;
@@ -794,35 +826,6 @@ int pw_impl_link_prepare(struct pw_impl_link *this)
 	return 0;
 }
 
-static int
-do_deactivate_link(struct spa_loop *loop,
-		   bool async, uint32_t seq, const void *data, size_t size, void *user_data)
-{
-        struct pw_impl_link *this = user_data;
-	struct impl *impl = SPA_CONTAINER_OF(this, struct impl, this);
-
-	pw_log_trace("%p: disable %p and %p", this, &this->rt.in_mix, &this->rt.out_mix);
-
-	spa_list_remove(&this->rt.out_mix.rt_link);
-	spa_list_remove(&this->rt.in_mix.rt_link);
-
-	if (this->input->node != this->output->node) {
-		struct pw_node_activation_state *state;
-
-		spa_list_remove(&this->rt.target.link);
-		state = &this->rt.target.activation->state[0];
-		if (this->rt.target.active) {
-			state->required--;
-			this->rt.target.active = false;
-		}
-
-		pw_log_trace("%p: node:%p state:%p pending:%d/%d", this, impl->inode,
-				state, state->pending, state->required);
-	}
-
-	return 0;
-}
-
 int pw_impl_link_deactivate(struct pw_impl_link *this)
 {
 	struct impl *impl = SPA_CONTAINER_OF(this, struct impl, this);
@@ -832,27 +835,26 @@ int pw_impl_link_deactivate(struct pw_impl_link *this)
 	if (!impl->activated)
 		return 0;
 
-	pw_loop_invoke(this->output->node->data_loop,
-		       do_deactivate_link, SPA_ID_INVALID, NULL, 0, true, this);
-
 	port_set_io(this, this->output, SPA_IO_Buffers, NULL, 0,
 			&this->rt.out_mix);
 	port_set_io(this, this->input, SPA_IO_Buffers, NULL, 0,
 			&this->rt.in_mix);
 
-	impl->io_set = false;
 	impl->activated = false;
 	pw_log_info("(%s) deactivated", this->name);
-	link_update_state(this, PW_LINK_STATE_PAUSED, 0, NULL);
 
+	if (this->info.state < PW_LINK_STATE_PAUSED || this->destroyed)
+		link_update_state(this, PW_LINK_STATE_INIT, 0, NULL);
+	else
+		link_update_state(this, PW_LINK_STATE_PAUSED, 0, NULL);
 	return 0;
 }
 
 static int
-global_bind(void *_data, struct pw_impl_client *client, uint32_t permissions,
+global_bind(void *object, struct pw_impl_client *client, uint32_t permissions,
 	       uint32_t version, uint32_t id)
 {
-	struct pw_impl_link *this = _data;
+	struct pw_impl_link *this = object;
 	struct pw_global *global = this->global;
 	struct pw_resource *resource;
 
@@ -966,6 +968,14 @@ static void input_port_latency_changed(void *data)
 		pw_impl_port_recalc_latency(this->output);
 }
 
+static void input_port_tag_changed(void *data)
+{
+	struct impl *impl = data;
+	struct pw_impl_link *this = &impl->this;
+	if (!this->feedback)
+		pw_impl_port_recalc_tag(this->output);
+}
+
 static void output_port_latency_changed(void *data)
 {
 	struct impl *impl = data;
@@ -974,11 +984,20 @@ static void output_port_latency_changed(void *data)
 		pw_impl_port_recalc_latency(this->input);
 }
 
+static void output_port_tag_changed(void *data)
+{
+	struct impl *impl = data;
+	struct pw_impl_link *this = &impl->this;
+	if (!this->feedback)
+		pw_impl_port_recalc_tag(this->input);
+}
+
 static const struct pw_impl_port_events input_port_events = {
 	PW_VERSION_IMPL_PORT_EVENTS,
 	.param_changed = input_port_param_changed,
 	.state_changed = input_port_state_changed,
 	.latency_changed = input_port_latency_changed,
+	.tag_changed = input_port_tag_changed,
 };
 
 static const struct pw_impl_port_events output_port_events = {
@@ -986,6 +1005,7 @@ static const struct pw_impl_port_events output_port_events = {
 	.param_changed = output_port_param_changed,
 	.state_changed = output_port_state_changed,
 	.latency_changed = output_port_latency_changed,
+	.tag_changed = output_port_tag_changed,
 };
 
 static void node_result(struct impl *impl, void *obj,
@@ -999,8 +1019,12 @@ static void input_node_result(void *data, int seq, int res, uint32_t type, const
 {
 	struct impl *impl = data;
 	struct pw_impl_port *port = impl->this.input;
-	pw_log_trace("%p: input port %p result seq:%d res:%d type:%u",
-			impl, port, seq, res, type);
+	pw_log_trace("%p: input port %p result seq:%d %d res:%d type:%u",
+			impl, port, seq, SPA_RESULT_ASYNC_SEQ(seq), res, type);
+
+	if (type == SPA_RESULT_TYPE_NODE_ERROR && impl->input_pending_seq == seq)
+		impl->input_result = res;
+
 	node_result(impl, &impl->this.input_link, seq, res, type, result);
 }
 
@@ -1008,8 +1032,12 @@ static void output_node_result(void *data, int seq, int res, uint32_t type, cons
 {
 	struct impl *impl = data;
 	struct pw_impl_port *port = impl->this.output;
-	pw_log_trace("%p: output port %p result seq:%d res:%d type:%u",
-			impl, port, seq, res, type);
+	pw_log_trace("%p: output port %p result seq:%d %d res:%d type:%u",
+			impl, port, seq, SPA_RESULT_ASYNC_SEQ(seq), res, type);
+
+	if (type == SPA_RESULT_TYPE_NODE_ERROR && impl->output_pending_seq == seq)
+		impl->output_result = res;
+
 	node_result(impl, &impl->this.output_link, seq, res, type, result);
 }
 
@@ -1108,18 +1136,84 @@ static void try_unlink_controls(struct impl *impl, struct pw_impl_port *output,
 	}
 }
 
+static int check_owner_permissions(struct pw_context *context,
+		struct pw_impl_node *node, struct pw_global *other, uint32_t permissions)
+{
+	const char *str;
+	struct pw_impl_client *client;
+	struct pw_global *global;
+	uint32_t perms;
+	uint32_t client_id;
+
+	str = pw_properties_get(node->properties, PW_KEY_CLIENT_ID);
+	if (str == NULL)
+		/* node not owned by client */
+		return 0;
+
+	if (!spa_atou32(str, &client_id, 0))
+		/* invalid client_id, something is wrong */
+		return -EIO;
+	if ((global = pw_context_find_global(context, client_id)) == NULL)
+		/* current client can't see the owner client */
+		return -errno;
+	if (!pw_global_is_type(global, PW_TYPE_INTERFACE_Client) ||
+	    (client = global->object) == NULL)
+		/* not the right object, something wrong */
+		return -EIO;
+
+	perms = pw_global_get_permissions(other, client);
+	if ((perms & permissions) != permissions)
+		/* owner client can't see other node */
+		return -EPERM;
+
+	return 0;
+}
+
 static int
 check_permission(struct pw_context *context,
 		 struct pw_impl_port *output,
 		 struct pw_impl_port *input,
 		 struct pw_properties *properties)
 {
+	int res;
+	uint32_t in_perms, out_perms;
+	struct pw_global *in_global, *out_global;
+
+	if ((in_global = input->node->global) == NULL)
+		return -ENOENT;
+	if ((out_global = output->node->global) == NULL)
+		return -ENOENT;
+
+	in_perms = out_perms = PW_PERM_R | PW_PERM_L;
+	if (context->current_client != NULL) {
+		in_perms = pw_global_get_permissions(in_global, context->current_client);
+		out_perms = pw_global_get_permissions(out_global, context->current_client);
+	}
+	/* current client can't see input node or output node */
+	if (!PW_PERM_IS_R(in_perms) || !PW_PERM_IS_R(out_perms))
+		return -ENOENT;
+
+	if ((res = check_owner_permissions(context, output->node,
+					in_global, PW_PERM_R)) < 0) {
+		/* output node owner can't see input node, check if the current
+		 * client has universal link permissions for the output node */
+		if (!PW_PERM_IS_L(out_perms))
+			return res;
+	}
+	if ((res = check_owner_permissions(context, input->node,
+					out_global, PW_PERM_R)) < 0) {
+		/* input node owner can't see output node, check if the current
+		 * client has universal link permissions for the input node */
+		if (!PW_PERM_IS_L(in_perms))
+			return res;
+	}
 	return 0;
 }
 
 static void permissions_changed(struct pw_impl_link *this, struct pw_impl_port *other,
 		struct pw_impl_client *client, uint32_t old, uint32_t new)
 {
+	int res;
 	uint32_t perm;
 
 	perm = pw_global_get_permissions(other->global, client);
@@ -1127,17 +1221,36 @@ static void permissions_changed(struct pw_impl_link *this, struct pw_impl_port *
 	new &= perm;
 	pw_log_debug("%p: permissions changed %08x -> %08x", this, old, new);
 
-	if (check_permission(this->context, this->output, this->input, this->properties) < 0) {
+	if ((res = check_permission(this->context, this->output, this->input, this->properties)) < 0) {
+		pw_log_info("%p: link permissions removed: %s", this, spa_strerror(res));
 		pw_impl_link_destroy(this);
 	} else if (this->global != NULL) {
 		pw_global_update_permissions(this->global, client, old, new);
 	}
 }
 
+static bool is_port_owner(struct pw_impl_client *client, struct pw_impl_port *port)
+{
+	const char *str;
+	uint32_t client_id;
+
+	str = pw_properties_get(port->node->properties, PW_KEY_CLIENT_ID);
+	if (str == NULL)
+		return false;
+
+	if (!spa_atou32(str, &client_id, 0))
+		return false;
+
+	return client_id == client->info.id;
+}
+
 static void output_permissions_changed(void *data,
 		struct pw_impl_client *client, uint32_t old, uint32_t new)
 {
 	struct pw_impl_link *this = data;
+	if (!is_port_owner(client, this->output) &&
+	    !is_port_owner(client, this->input))
+		return;
 	permissions_changed(this, this->input, client, old, new);
 }
 
@@ -1150,6 +1263,9 @@ static void input_permissions_changed(void *data,
 		struct pw_impl_client *client, uint32_t old, uint32_t new)
 {
 	struct pw_impl_link *this = data;
+	if (!is_port_owner(client, this->output) &&
+	    !is_port_owner(client, this->input))
+		return;
 	permissions_changed(this, this->output, client, old, new);
 }
 
@@ -1169,6 +1285,7 @@ struct pw_impl_link *pw_context_create_link(struct pw_context *context,
 	struct impl *impl;
 	struct pw_impl_link *this;
 	struct pw_impl_node *input_node, *output_node;
+	const char *str;
 	int res;
 
 	if (output == input)
@@ -1209,8 +1326,6 @@ struct pw_impl_link *pw_context_create_link(struct pw_context *context,
                 this->user_data = SPA_PTROFF(impl, sizeof(struct impl), void);
 
 	impl->work = pw_context_get_work_queue(context);
-	if (impl->work == NULL)
-		goto error_work_queue;
 
 	this->context = context;
 	this->properties = properties;
@@ -1220,7 +1335,20 @@ struct pw_impl_link *pw_context_create_link(struct pw_context *context,
 	this->input = input;
 
 	/* passive means that this link does not make the nodes active */
-	this->passive = pw_properties_get_bool(properties, PW_KEY_LINK_PASSIVE, false);
+	str = pw_properties_get(properties, PW_KEY_LINK_PASSIVE);
+	this->passive = str ? spa_atob(str) :
+		(output->passive && input_node->can_suspend) ||
+		(input->passive && output_node->can_suspend) ||
+		(input->passive && output->passive);
+	if (this->passive && str == NULL)
+		 pw_properties_set(properties, PW_KEY_LINK_PASSIVE, "true");
+
+	impl->async = (output_node->async || input_node->async) &&
+		SPA_FLAG_IS_SET(output->flags, PW_IMPL_PORT_FLAG_ASYNC) &&
+		SPA_FLAG_IS_SET(input->flags, PW_IMPL_PORT_FLAG_ASYNC);
+
+	if (impl->async)
+		 pw_properties_set(properties, PW_KEY_LINK_ASYNC, "true");
 
 	spa_hook_list_init(&this->listener_list);
 
@@ -1251,8 +1379,6 @@ struct pw_impl_link *pw_context_create_link(struct pw_context *context,
 	spa_list_append(&output->links, &this->output_link);
 	spa_list_append(&input->links, &this->input_link);
 
-	impl->io = SPA_IO_BUFFERS_INIT;
-
 	select_io(this);
 
 	if (this->feedback) {
@@ -1264,18 +1390,17 @@ struct pw_impl_link *pw_context_create_link(struct pw_context *context,
 		impl->inode = input_node;
 	}
 
-	this->rt.target.signal = impl->inode->rt.target.signal;
-	this->rt.target.data = impl->inode->rt.target.data;
-
 	pw_log_debug("%p: constructed out:%p:%d.%d -> in:%p:%d.%d", impl,
 		     output_node, output->port_id, this->rt.out_mix.port.port_id,
 		     input_node, input->port_id, this->rt.in_mix.port.port_id);
 
-	if (asprintf(&this->name, "%d.%d -> %d.%d",
-			output_node->info.id, output->port_id,
-			input_node->info.id, input->port_id) < 0)
-		this->name = NULL;
-	pw_log_info("(%s) (%s) -> (%s)", this->name, output_node->name, input_node->name);
+	this->name = spa_aprintf("%d.%d.%d -> %d.%d.%d",
+			output_node->info.id, output->port_id, this->rt.out_mix.port.port_id,
+			input_node->info.id, input->port_id, this->rt.in_mix.port.port_id);
+	pw_log_info("(%s) (%s) -> (%s) async:%d:%d:%d:%04x:%04x:%d", this->name, output_node->name,
+			input_node->name, output_node->driving,
+			output_node->async, input_node->async,
+			output->flags, input->flags, impl->async);
 
 	pw_impl_port_emit_link_added(output, this);
 	pw_impl_port_emit_link_added(input, this);
@@ -1284,8 +1409,11 @@ struct pw_impl_link *pw_context_create_link(struct pw_context *context,
 
 	pw_impl_port_recalc_latency(output);
 	pw_impl_port_recalc_latency(input);
+	pw_impl_port_recalc_tag(output);
+	pw_impl_port_recalc_tag(input);
 
-	pw_impl_node_emit_peer_added(impl->onode, impl->inode);
+	if (impl->onode != impl->inode)
+		this->peer = pw_node_peer_ref(impl->onode, impl->inode);
 
 	return this;
 
@@ -1309,10 +1437,6 @@ error_no_mem:
 	res = -errno;
 	pw_log_debug("alloc failed: %m");
 	goto error_exit;
-error_work_queue:
-	res = -errno;
-	pw_log_debug("work queue failed: %m");
-	goto error_free;
 error_output_mix:
 	pw_log_error("%p: can't get output mix %d (%s)", this, res, spa_strerror(res));
 	goto error_free;
@@ -1328,9 +1452,9 @@ error_exit:
 	return NULL;
 }
 
-static void global_destroy(void *object)
+static void global_destroy(void *data)
 {
-	struct pw_impl_link *link = object;
+	struct pw_impl_link *link = data;
 	spa_hook_remove(&link->global_listener);
 	link->global = NULL;
 	pw_impl_link_destroy(link);
@@ -1346,7 +1470,6 @@ int pw_impl_link_register(struct pw_impl_link *link,
 		     struct pw_properties *properties)
 {
 	static const char * const keys[] = {
-		PW_KEY_OBJECT_SERIAL,
 		PW_KEY_OBJECT_PATH,
 		PW_KEY_MODULE_ID,
 		PW_KEY_FACTORY_ID,
@@ -1375,6 +1498,7 @@ int pw_impl_link_register(struct pw_impl_link *link,
 	link->global = pw_global_new(context,
 				     PW_TYPE_INTERFACE_Link,
 				     PW_VERSION_LINK,
+				     PW_LINK_PERM_MASK,
 				     properties,
 				     global_bind,
 				     link);
@@ -1414,9 +1538,12 @@ SPA_EXPORT
 void pw_impl_link_destroy(struct pw_impl_link *link)
 {
 	struct impl *impl = SPA_CONTAINER_OF(link, struct impl, this);
+	bool was_prepared = link->prepared;
 
 	pw_log_debug("%p: destroy", impl);
 	pw_log_info("(%s) destroy", link->name);
+
+	link->destroyed = true;
 	pw_impl_link_emit_destroy(link);
 
 	pw_impl_link_deactivate(link);
@@ -1424,7 +1551,8 @@ void pw_impl_link_destroy(struct pw_impl_link *link)
 	if (link->registered)
 		spa_list_remove(&link->link);
 
-	pw_impl_node_emit_peer_removed(impl->onode, impl->inode);
+	if (link->peer)
+		pw_node_peer_unref(link->peer);
 
 	try_unlink_controls(impl, link->output, link->input);
 
@@ -1436,7 +1564,7 @@ void pw_impl_link_destroy(struct pw_impl_link *link)
 		pw_global_destroy(link->global);
 	}
 
-	if (link->prepared)
+	if (was_prepared)
 		pw_context_recalc_graph(link->context, "link destroy");
 
 	pw_log_debug("%p: free", impl);
@@ -1450,6 +1578,7 @@ void pw_impl_link_destroy(struct pw_impl_link *link)
 
 	free(link->name);
 	free(link->info.format);
+	free((char *) link->info.error);
 	free(impl);
 }
 
